from .constants import TaskPriority, TaskState
from .consumer import TaskConsumer, TaskManager
from .crontab import crontab
from .every import every
from .scheduler import TaskScheduler
from .task import (
    Task,
    TaskConstructor,
    TaskContext,
    TaskDecoratorError,
    TaskExecutor,
    TaskRunError,
    create_task_app,
    task,
)
from .task_run import TaskRun

__all__ = [
    "TaskPriority",
    "TaskState",
    "TaskScheduler",
    "task",
    "Task",
    "TaskContext",
    "TaskRunError",
    "TaskExecutor",
    "TaskConstructor",
    "TaskDecoratorError",
    "TaskManager",
    "TaskConsumer",
    "TaskRun",
    "create_task_app",
    "crontab",
    "every",
]



import os
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Callable, Dict, Iterable, NamedTuple, Optional, Tuple, Union
from uuid import uuid4

from yarl import URL

from fluid import json
from fluid.redis import FluidRedis
from fluid.utils import microseconds

from .constants import TaskPriority, TaskState
from .task import Task
from .task_run import TaskRun

_brokers = {}

DEFAULT_BROKER_URL = "redis://localhost:6379/3"


class UnknownTask(RuntimeError):
    pass


class TaskRegistry(Dict[str, Task]):
    def periodic(self) -> Iterable[Task]:
        for task in self.values():
            if task.schedule:
                yield task


class QueuedTask(NamedTuple):
    run_id: str
    task: str
    priority: Optional[TaskPriority]
    params: Dict[str, Any]


class Broker(ABC):
    def __init__(self, url: URL) -> None:
        self.url: URL = url
        self.registry: TaskRegistry = TaskRegistry()

    @abstractmethod
    async def queue_task(self, queued_task: QueuedTask) -> TaskRun:
        """Queue a task"""

    @abstractmethod
    async def get_task_run(self) -> Optional[TaskRun]:
        """Get a Task run from the task queue"""

    @abstractmethod
    async def queue_length(self) -> Dict[str, int]:
        """Get a Task run from the task queue"""

    async def close(self) -> None:
        """Close the broker on shutdown"""

    def new_uuid(self) -> str:
        return uuid4().hex

    def task_from_registry(self, task: Union[str, Task]) -> Task:
        if isinstance(task, Task):
            self.register_task(task)
        else:
            task_ = self.registry.get(task)
            if not task_:
                raise UnknownTask(task)
            task = task_
        return task

    def register_task(self, task: Task) -> None:
        self.registry[task.name] = task

    def task_run_from_data(self, data: Dict[str, Any]) -> TaskRun:
        """Build a TaskRun object from its metadata"""
        data = data.copy()
        name = data.pop("name")
        data["task"] = self.task_from_registry(name)
        return TaskRun(**data)

    def task_run_data(
        self, queued_task: QueuedTask, state: TaskState
    ) -> Dict[str, Any]:
        """Create a dictionary of metadata required by a task run

        This dictionary must be serializable by the broker
        """
        task = self.task_from_registry(queued_task.task)
        priority = queued_task.priority or task.priority
        return dict(
            id=queued_task.run_id,
            name=task.name,
            priority=priority.name,
            state=state.name,
            params=queued_task.params,
            queued=microseconds(),
        )

    @classmethod
    def from_url(cls, url: str = "") -> "Broker":
        url = url or os.getenv("SCHEDULER_BROKER_URL", DEFAULT_BROKER_URL)
        p = URL(url)
        Factory = _brokers.get(p.scheme)
        if not Factory:
            raise RuntimeError(f"Invalid broker {url}")
        return Factory(p)

    @classmethod
    def register_broker(cls, name: str, factory: Callable):
        _brokers[name] = factory


class RedisBroker(Broker):
    """A simple broker based on redis lists"""

    @cached_property
    def redis(self) -> FluidRedis:
        return FluidRedis(str(self.url.with_query({})), name=self.name)

    @property
    def name(self) -> str:
        return self.url.query.get("name", "redis-task-broker")

    @cached_property
    def task_queue_names(self) -> Tuple[str, ...]:
        return tuple(self.task_queue_name(p) for p in TaskPriority)

    def task_queue_name(self, priority: TaskPriority) -> str:
        return f"{self.name}-queue-{priority.name}"

    async def queue_length(self) -> Dict[str, int]:
        pipe = self.redis.cli.pipeline()
        for name in self.task_queue_names:
            pipe.llen(name)
        result = await pipe.execute()
        return {p.name: r for p, r in zip(TaskPriority, result)}

    async def close(self) -> None:
        """Close the broker on shutdown"""
        await self.redis.close()

    async def get_task_run(self) -> Optional[TaskRun]:
        data = await self.redis.cli.brpop(self.task_queue_names, timeout=1)
        if data:
            data_str = data[1].decode("utf-8")
            return self.task_run_from_data(json.loads(data_str))
        return None

    async def queue_task(self, queued_task: QueuedTask) -> TaskRun:
        task = self.task_from_registry(queued_task.task)
        priority = queued_task.priority or task.priority
        data = self.task_run_data(queued_task, TaskState.queued)
        await self.redis.cli.lpush(self.task_queue_name(priority), json.dumps(data))
        return self.task_run_from_data(data)


Broker.register_broker("redis", RedisBroker)



import asyncio
import logging
from collections import defaultdict, deque
from functools import cached_property
from typing import Callable, Dict, NamedTuple, Optional, Union

from inflection import underscore

from fluid.node import Consumer, NodeWorkers, Worker
from fluid.utils import microseconds

from .broker import Broker, QueuedTask, TaskRegistry, UnknownTask
from .constants import TaskPriority, TaskState
from .task import Task
from .task_run import TaskRun
from .utils import WaitFor

ConsumerCallback = Callable[[TaskRun, "TaskManager"], None]


class TaskFailure(RuntimeError):
    pass


class Event(NamedTuple):
    type: str
    tag: str

    @classmethod
    def from_string(cls, event: str) -> "Event":
        bits = event.split(".")
        return cls(bits[0], bits[1] if len(bits) > 1 else "")


class TaskManager(NodeWorkers):
    """Base class for both TaskConsumer and TaskScheduler"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._msg_handlers: Dict[str, Dict[str, ConsumerCallback]] = defaultdict(dict)
        self._queue_tasks_worker = Worker(
            self._queue_tasks, logger=self.logger.getChild("queue")
        )

    @cached_property
    def broker(self) -> Broker:
        return Broker.from_url()

    @property
    def registry(self) -> TaskRegistry:
        return self.broker.registry

    @cached_property
    def task_queue(self) -> asyncio.Queue:
        return asyncio.Queue()

    @property
    def type(self) -> str:
        return underscore(self.__class__.__name__)

    async def setup(self) -> None:
        self._consumer = Consumer(
            self._execute_async, logger=self.logger.getChild("internal-consumer")
        )
        self.add_workers(self._consumer)
        await self._queue_tasks_worker.start_app(self.app)

    async def teardown(self) -> None:
        await self._queue_tasks_worker.close()
        await self.broker.close()

    def register_task(self, task: Task) -> None:
        """Register a task with the task manager

        Only tasks registered can be executed by a task manager
        """
        self.broker.register_task(task)

    def queue(
        self,
        task: str,
        priority: Optional[TaskPriority] = None,
        **params,
    ) -> QueuedTask:
        """Queue a Task for execution and return the QueuedTask object"""
        queued_task = QueuedTask(
            run_id=self.broker.new_uuid(),
            task=task,
            priority=priority,
            params=params,
        )
        self.task_queue.put_nowait(queued_task)
        return queued_task

    def dispatch(self, task_run: TaskRun, event_type: str) -> str:
        """Dispatch a message to the registered handlers.

        It returns the asset affected by the message (if any)
        """
        handlers = self._msg_handlers.get(event_type)
        if handlers:
            for handler in handlers.values():
                handler(task_run, self)

    def register_handler(self, event: str, handler: ConsumerCallback) -> None:
        event = Event.from_string(event)
        self._msg_handlers[event.type][event.tag] = handler

    def unregister_handler(self, event: str) -> None:
        event = Event.from_string(event)
        self._msg_handlers[event.type].pop(event.tag, None)

    def execute_async(self, async_callable, *args) -> None:
        self._consumer.submit((async_callable, args))

    # process tasks from the internal queue
    async def _queue_tasks(self) -> None:
        while self.is_running():
            queued_task = await self.task_queue.get()
            task_run = await self.broker.queue_task(queued_task)
            self.dispatch(task_run, "queued")
            await asyncio.sleep(0)

    async def _execute_async(self, message) -> None:
        try:
            executable, args = message
            await executable(*args)
        except Exception:
            self.logger.exception("unhandled exception while executing async callaback")


class ConsumerConfig(NamedTuple):
    max_concurrent_tasks: int = 10
    """number of coroutine workers"""
    sleep: float = 0.1
    """amount to sleep after completion of a task"""


class TaskConsumer(TaskManager):
    def __init__(self, **config) -> None:
        super().__init__()
        self.cfg: ConsumerConfig = ConsumerConfig(**config)
        self._concurrent_tasks: Dict[str, Dict[str, TaskRun]] = defaultdict(dict)
        self._priority_task_run_queue = deque()
        for i in range(self.cfg.max_concurrent_tasks):
            self.add_workers(
                Worker(
                    self._consume_tasks, logger=self.logger.getChild(f"worker-{i+1}")
                )
            )

    @property
    def num_concurrent_tasks(self) -> int:
        """The number of concurrent_tasks"""
        return sum(len(v) for v in self._concurrent_tasks.values())

    def num_concurrent_tasks_for(self, task_name) -> int:
        """The number of concurrent_tasks"""
        return len(self._concurrent_tasks[task_name])

    async def queue_and_wait(self, task: Union[str, Task], **params) -> TaskRun:
        run_id = self.execute(task, **params).id
        waitfor = WaitFor(run_id=run_id)
        self.register_handler(f"end.{waitfor.run_id}", waitfor)
        try:
            return await waitfor.waiter
        finally:
            self.unregister_handler(f"end.{waitfor.run_id}")

    def execute(self, task: Union[Task, str], **params) -> TaskRun:
        """Execute a Task by-passing the broker task queue"""
        queued_task = QueuedTask(
            run_id=self.broker.new_uuid(),
            task=task,
            priority="",
            params=params,
        )
        data = self.broker.task_run_data(queued_task, TaskState.queued)
        task_run = self.broker.task_run_from_data(data)
        self._priority_task_run_queue.appendleft(task_run)
        return task_run

    # Internals
    async def _consume_tasks(self) -> None:
        while self.is_running():
            if self._priority_task_run_queue:
                task_run = self._priority_task_run_queue.pop()
            else:
                try:
                    task_run = await self.broker.get_task_run()
                except UnknownTask as exc:
                    self.logger.error(
                        "unknown task %s - it looks like it is not "
                        "registered with this consumer",
                        exc,
                    )
                    task_run = None
                if not task_run:
                    await asyncio.sleep(self.cfg.sleep)
                    continue
            task_name = task_run.name
            task_run.start = microseconds()
            task_run.set_state(TaskState.running)
            task_context = task_run.task.create_context(self, task_run=task_run)
            #
            if task_run.task.max_concurrency <= self.num_concurrent_tasks_for(
                task_name
            ):
                task_run.set_state(TaskState.rate_limited)
                task_run.waiter.set_result(None)
            else:
                task_context.logger.info("start")
                self._concurrent_tasks[task_name][task_run.id] = task_run
                self.dispatch(task_run, "start")
                try:
                    result = await task_run.task.executor(task_context)
                except Exception as exc:
                    task_run.waiter.set_exception(exc)
                    task_run.set_state(TaskState.failure)
                else:
                    if task_run.is_failure:
                        task_run.waiter.set_exception(TaskFailure())
                    else:
                        task_run.waiter.set_result(result)
                        if not task_run.in_finish_state:
                            task_run.set_state(TaskState.success)
            try:
                await task_run.waiter
            except TaskFailure:
                # task_context.logger.warning("CPU bound task failure")
                # no need to log here, the log is already done from the CPU bound script
                pass
            except Exception:
                task_context.logger.exception("critical exception while executing")
            task_run.end = microseconds()
            self._concurrent_tasks[task_name].pop(task_run.id, None)
            self.dispatch(task_run, "end")
            task_context.logger.log(
                logging.WARNING if task_run.is_failure else logging.INFO,
                "end - %s - milliseconds - %s",
                task_run.state,
                round(0.001 * task_run.duration, 3),
            )
            await asyncio.sleep(self.cfg.sleep)



from contextlib import asynccontextmanager

from aiohttp.test_utils import TestClient, TestServer
from aiohttp.web import Application
from openapi.json import dumps


@asynccontextmanager
async def app_cli(app: Application) -> TestClient:
    server = TestServer(app)
    client = TestClient(server, json_serialize=dumps)
    await client.start_server()
    try:
        yield client
    finally:
        await client.close()



from datetime import datetime, timedelta
from typing import Optional

from .crontab import CronRun, ScheduleType


def every(delta: timedelta) -> ScheduleType:
    def validate(
        timestamp: datetime, last_run: Optional[CronRun] = None
    ) -> Optional[CronRun]:
        year, month, day, hour, minute, second, _, _, _ = timestamp.timetuple()
        run = CronRun(year, month, day, hour, minute, second)
        if not last_run or timestamp - last_run.datetime >= delta:
            return run
        return None

    return validate



"""Originally from https://github.com/coleifer/huey
"""
import re
from datetime import datetime
from typing import Callable, NamedTuple, Optional, Union

dash_re = re.compile(r"(\d+)-(\d+)")
every_re = re.compile(r"\*\/(\d+)")
CI = Union[str, int]


class CronRun(NamedTuple):
    year: int
    month: int
    day: int
    week_day: int
    hour: int
    minute: int
    second: int = 0

    @property
    def datetime(self) -> datetime:
        return datetime(*self)


ScheduleType = Callable[[datetime, CronRun], Optional[CronRun]]


def crontab(
    minute: CI = "*",
    hour: CI = "*",
    day: CI = "*",
    month: CI = "*",
    day_of_week: CI = "*",
) -> ScheduleType:
    """
    Convert a "crontab"-style set of parameters into a test function that will
    return True when the given datetime matches the parameters set forth in
    the crontab.
    For day-of-week, 0=Sunday and 6=Saturday.
    Acceptable inputs:
    * = every distinct value
    */n = run every "n" times, i.e. hours='*/4' == 0, 4, 8, 12, 16, 20
    m-n = run every time m..n
    m,n = run on m and n
    """
    validation = (
        ("m", month, range(1, 13)),
        ("d", day, range(1, 32)),
        ("w", day_of_week, range(8)),  # 0-6, but also 7 for Sunday.
        ("H", hour, range(24)),
        ("M", minute, range(60)),
    )
    cron_settings = []

    for (date_str, value, acceptable) in validation:
        settings = set()

        if isinstance(value, int):
            value = str(value)

        for piece in value.split(","):
            if piece == "*":
                settings.update(acceptable)
                continue

            if piece.isdigit():
                piece = int(piece)
                if piece not in acceptable:
                    raise ValueError("%d is not a valid input" % piece)
                elif date_str == "w":
                    piece %= 7
                settings.add(piece)

            else:
                dash_match = dash_re.match(piece)
                if dash_match:
                    lhs, rhs = map(int, dash_match.groups())
                    if lhs not in acceptable or rhs not in acceptable:
                        raise ValueError("%s is not a valid input" % piece)
                    elif date_str == "w":
                        lhs %= 7
                        rhs %= 7
                    settings.update(range(lhs, rhs + 1))
                    continue

                # Handle stuff like */3, */6.
                every_match = every_re.match(piece)
                if every_match:
                    if date_str == "w":
                        raise ValueError(
                            "Cannot perform this kind of matching" " on day-of-week."
                        )
                    interval = int(every_match.groups()[0])
                    settings.update(acceptable[::interval])

        cron_settings.append(sorted(list(settings)))

    cron_settings = tuple(cron_settings)

    def validate_date(
        timestamp: datetime, last_run: Optional[CronRun] = None
    ) -> Optional[CronRun]:
        year, month, day, hour, minute, _, w, _, _ = timestamp.timetuple()
        run = CronRun(year, month, day, (w + 1) % 7, hour, minute)
        if last_run == run:
            return None

        for (date_piece, selection) in zip(run[1:], cron_settings):
            if date_piece not in selection:
                return None
        return run

    return validate_date



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fluid', 'fluid.github', 'fluid.scheduler']

package_data = \
{'': ['*']}

install_requires = \
['aio-kong>=0.9.0,<0.10.0',
 'aio-openapi>=2.6.0,<3.0.0',
 'aiobotocore[boto3]>=2.0.1,<3.0.0',
 'aioconsole>=0.3.1,<0.4.0',
 'aiohttp_cors>=0.7.0,<0.8.0',
 'aioredis>=2.0.0,<3.0.0',
 'colorlog>=6.6.0,<7.0.0',
 'inflection>=0.5.1,<0.6.0',
 'prometheus-async>=19.2.0,<20.0.0',
 'pycountry>=20.7.3,<21.0.0',
 'python-json-logger>=2.0.2,<3.0.0',
 'python-slugify[unidecode]>=5.0.2,<6.0.0',
 's3fs>=2021.11.1,<2022.0.0',
 'ujson>=5.1.0,<6.0.0',
 'uvloop>=0.16.0,<0.17.0']

setup_kwargs = {
    'name': 'aio-fluid',
    'version': '0.3.2',
    'description': 'Tools for backend python services',
    'long_description': '# Tools for backend python services\n\n[![PyPI version](https://badge.fury.io/py/aio-fluid.svg)](https://badge.fury.io/py/aio-fluid)\n[![build](https://github.com/quantmind/fluid/workflows/build/badge.svg)](https://github.com/quantmind/aio-fluid/actions?query=workflow%3Abuild)\n[![codecov](https://codecov.io/gh/quantmind/aio-fluid/branch/master/graph/badge.svg?token=81oWUoyEVp)](https://codecov.io/gh/quantmind/aio-fluid)\n\n## Installation\n\nThis is a simple python package you can install via pip:\n\n```\npip install aio-fluid\n```\n\n## Modules\n\n### [scheduler](./fluid/scheduler)\n\nA simple asynchronous task queue with a scheduler\n\n### [kernel](./fluid/kernel)\n\nAsync utility for executing commands in sub-processes\n\n## AWS\n\npackages for AWS interaction are installed via\n\n- [aiobotocore](https://github.com/aio-libs/aiobotocore)\n- [s3fs](https://github.com/fsspec/s3fs)(which depends on aiobotocore and therefore versions must be compatible)\n- [boto3](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html) is installed as extra dependency of aiobotocore so versioning is compatible\n',
    'author': 'Luca',
    'author_email': 'luca@quantmind.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)
